/****************************************************************************/
/*                                 RCOPY                                    */
/*                                                                          */
/* DATE: 6/06/86                                                            */
/*                                                                          */
/* COPYRIGHT:                                                               */
/*               AML/2 Application Development Environment                  */
/*               7505-CAC (C) Copyright IBM Corporation 1986                */
/*               Licensed Material - Program Property of IBM                */
/*                                                                          */
/* Purpose: RCOPY is used to either upload or download a file to a 7575/6   */
/*          manufacturing control system.                                   */
/*                                                                          */
/* Format:  Download - RCOPY sourcefile R:[targetfile] [/D:device][/T:#][/P]*/
/*          Upload   - RCOPY R:targetfile [sourcefile] [/D:device][/T:#][/P]*/
/*                                                                          */
/* Options:     /D:device - Specifies the device driver.  Should be either  */
/*                          /D:ROLLUP1 or /D:ROLLUP2.  Default is ROLLUP1.  */
/*                                                                          */
/*              /T:#      - Specifies the timeout value.  Default is 5 secs.*/
/*                                                                          */
/*              /P        - Indicates that a dump of all transmited and     */
/*                          received lines be done to stdout.               */
/*                                                                          */
/*              If either the targetfile (for a download) or sourcefile     */
/*              (for an upload) are omitted, then the filename from the     */
/*              first file (with the path removed) is used.                 */
/*                                                                          */
/*  PROCESSOR/COMPILER USED:                                                */
/*      C Compiler                                                          */
/*      cc rcopy /DLINT_ARGS;                                               */
/*      clink rcopy;                                                        */
/*                                                                          */
/* Errors:                                                                  */
/* 0        Two files must be given.                                        */
/* 1        R: omitted.                                                     */
/* 2        Only one R: allowed.                                            */
/* 3        Unable to open source file.                                     */
/* 4        Unable to open target file.                                     */
/* 5        Unable to open comm port.                                       */
/* 6        Read/write error from comm port.                                */
/* 7        Communications timeout.                                         */
/* 8        Invalid parameter.                                              */
/* 9        Cannot load while program running.                              */
/* 10       reserved msg                                                    */
/* 11       File not successfully saved.                                    */
/*          Exception ???? occurred.                                        */
/*                                                                          */
/*  CHANGE ACTIVITY:                                                        */
/*                                                                          */
/*   DATE       SUBROUTINE CHANGED                                          */
/*  --------   -----------------------------------------------------        */
/*  4/24/86    LSM - don't create 0 byte files  on upload or send down files*/
/*              of 0 bytes on download- main(argc,argv)                     */
/*  4/30/86    ME - /t option overwite end of string marker                 */
/*             -commandline(argc,argv)                                      */
/*             LSM - make sure that the timeout value specified is at least */
/*              the DEF_TIMEOUT    -commandline(argc,argv)                  */
/*             LSM - rcopy r: a return msg that 2 files must be given       */
/*             -main(argc,argv)                                             */
/*  5/6/86     LSM - downloading to contoller DISKETTE msg change           */
/*             -main(argc,argv)                                             */
/*  7/1/86     MEE - Allow a NOTSAVED message on a download                 */
/*             -main(argc,argv)                                             */
/****************************************************************************/

#define BMODE           0x20
#define ISDEV           0x80
#define IOCTL           0x44

#define READFILE  'R'
#define SENDFILE  'S'
#define DATAREADY 'D'
#define GOODFILE  'G'
#define BADFILE   'B'
#define FILESAVED 'F'
#define NOTSAVED  'N'

#define BUFSIZE         256

#define H_READ "220"
#define H_WRITE "221"
#define H_STATUS "202"
#define H_STOP "204"
#define H_STARTRUN "205"
#define H_SELECT "320"
#define H_UPLOAD "400"
#define H_DOWNLOAD "401"
#define H_HOME "250"
#define DEF_TIMEOUT 10

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
#include <time.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>

/****************************************************************************/
/*                                                                          */
/*                      External Variables Owned by RCOPY.C                 */
/*                                                                          */
/****************************************************************************/
char recv[BUFSIZE];     /* The controller's recv to a message               */
char xmit[BUFSIZE];     /* A string to be transmitted to the controller     */
int traceflag=0;        /* Indicates if dumping to stdprn should be done    */
int handle;             /* file handle of device driver                     */
union REGS regs;        /* Used by IBM C to interface to DOS/BIOS.          */
int seq;                /* The sequence number used                         */
char device[9];         /* Holds the device driver to be used               */
int timeout=DEF_TIMEOUT;/* Holds the time out value                         */
char *errors[] = {
        "Two files must be given.",
        "R: omitted.",
        "Only one R: allowed.",
        "Unable to open source file.",
        "Unable to open target file.",
        "Unable to open comm port.",
        "Read/write error from comm port.",
        "Communications timeout.",
        "Invalid parameter.",
        "Cannot load while program running.",
        "reseved msg      .",
        "File not successfully saved."};

/****************************************************************************/
/* Procedure Goodbye  -                                                     */
/*      Exits the program, first printing an error.                         */
/****************************************************************************/
void goodbye(errnum)
        int errnum;
        {
        char *p;

        if (errnum < 12) fprintf(stderr,"%s\n",errors[errnum]);
                else {
                strtok(recv," '");    /* skip 0 or err#*/
                strtok(NULL," '");    /* skip seq */
                strtok(NULL," '");    /* skip code or 222*/
                p=strtok(NULL," '");  /* get exception */

                fprintf(stderr,"Exception %s occurred.\n",p);
                }
        /* LSM 7/18/86 code to let Bristol know that rollup is not attached */
        sendit(H_WRITE,"{ IOCTL(ROLLUP,3,@'0B00@'X);_ROLLUP=FALSE; }");

        exit(1);
        }


/****************************************************************************/
/* Procedure Numinbuffer -                                                  */
/*      Returns the number of characters ready to be read from device.      */
/*      0  ==>  if IOCTL error or no characters available.                  */
/*      >0 ==>  Number of characters ready to be read.                      */
/*      -1 ==>  EOT received (i.e. END OF FILE marker).                     */
/****************************************************************************/
static int numinbuffer() {
        struct {
                int num_received;
                int num_to_send;
                } q;

        /* Do IOCTL to find out values */
        regs.h.ah=IOCTL;
        regs.x.bx=handle;
        regs.h.al=2;
        regs.x.cx=4;
        regs.x.dx=(int)&q;

        intdos(&regs,&regs);
        if (regs.x.cflag == 1) return(0);

        return(q.num_received);
        }



/****************************************************************************/
/* Procedure Xloc  -                                                        */
/*      Returns the number of free slots available in the send buffer.      */
/*      0  ==>  if IOCTL error or no slots available.                       */
/*      >0 ==>  Number of slots available.                                  */
/****************************************************************************/
static int xloc() {
        struct {
                int num_received;
                int num_to_send;
                } q;

        /* Do IOCTL to find out values */
        regs.h.ah=IOCTL;
        regs.x.bx=handle;
        regs.h.al=2;
        regs.x.cx=4;
        regs.x.dx=(int)&q;

        intdos(&regs,&regs);
        if (regs.x.cflag == 1) return(0);

        return(q.num_to_send);
        }



/****************************************************************************/
/* Procedure Xmiteof -                                                      */
/*      Sends an EOT (i.e. END OF FILE marker) to the controller.           */
/****************************************************************************/
static void xmiteof() {
        int eot=5;              /* MACHINE DEPENDENT - Requires to
                                   be stored in memory as 0x0500 */
        /* Echo to stdprn if requested by user */
        if (traceflag) fprintf(stdprn,"XMIT: EOF\n");
        fflush(stdprn);

        /* Do IOCTL xmit EOT */
        regs.h.ah=IOCTL;
        regs.x.bx=handle;
        regs.h.al=3;
        regs.x.cx=2;
        regs.x.dx=(int)&eot;

        intdos(&regs,&regs);
        }


/****************************************************************************/
/* Procedure Tracewrite -                                                   */
/*      Performs a write, but first echos the values to the printer if      */
/*      traceflag is set.                                                   */
/* Parameters -                                                             */
/*      cnt       - The number of characters to be sent from xmit           */
/*      timeouterr- The error code to exit with should a timeout occur      */
/*      enddnld   - Indicates that the download protocol should be ended    */
/*                  (i.e. transmit EOF and BADFILE).                        */
/****************************************************************************/
tracewrite(cnt,timeouterr,enddnld)
        int cnt, timeouterr, enddnld;
        {
        int rc;
        long starttime, endtime;        /* used to implement timeout        */
        char *buf=xmit;                 /* points to characters xmitted     */

        /* Wait for space in XMIT buffer of device driver */
        time(&starttime);
        for (endtime=starttime;xloc() < cnt && endtime-starttime < timeout;time(&endtime) );

        /* If still no space, kill program */
        if (xloc() < cnt) {
                /* End download protocol if requested */
                if (enddnld) {
                        /* Send end of file marker */
                        xmiteof();

                        /* Send BADFILE message if there's space */
                        time(&starttime);
                        for (endtime=starttime;!xloc() && endtime-starttime < timeout;time(&endtime) );
                        if (xloc()) {
                                *xmit=BADFILE;
                                if (traceflag) fprintf(stderr,"XMIT: %c\n",*xmit);
                                write(handle,xmit,1);
                                }
                        }
                goodbye(timeouterr);
                }

        /* Send line to controller */
        if ((rc=write(handle,xmit,cnt)) < 0) {
                /* End download protocol if requested */
                if (enddnld) {
                        /* Send end of file marker */
                        xmiteof();

                        /* Send BADFILE message if there's space */
                        time(&starttime);
                        for (endtime=starttime;!xloc() && endtime-starttime < timeout;time(&endtime) );
                        if (xloc()) {
                                *xmit=BADFILE;
                                if (traceflag) fprintf(stderr,"XMIT: %c\n",*xmit);
                                write(handle,xmit,1);
                                }
                        }
                goodbye(6);
                }

        if (traceflag) {
                fprintf(stdprn,"XMIT: ");
                for (cnt=rc;cnt--;fputc(*buf++,stdprn)) ;
                fputc('\n',stdprn);
                fflush(stdprn);
                }
        return(rc);
        }

/****************************************************************************/
/* Procedure Openport -                                                     */
/*      Tries to open the specified device driver.  Returns 0 if ok, -1 if  */
/*      error.                                                              */
/****************************************************************************/
static int openport(device)
        char *device;           /* Name of the device driver */
        {
        int ddata=3;            /* MACHINE DEPENDENT - Requires to
                                   be stored in memory as 0x0300 */


        /* Try to open device and return -1 if error */
        if ((handle=open(device,O_RDWR|O_BINARY)) == -1) return(-1);

        /* Must first tell DOS that device driver is binary */

        /* First do function 0, get control word.  See IOCTL description
           in DOS Tech. Ref. */
        regs.h.ah=IOCTL;
        regs.x.bx=handle;
        regs.h.al=0;
        intdos(&regs,&regs);
        if (regs.x.cflag == 1) {close(handle); return(-1);}

        /* If a file, return error */
        if (!(regs.h.dl & ISDEV)) {close(handle); return(-1);}

        /* Next or the binary mode flag of the control word and write
           this back to DOS using an IOCTL. */
        regs.h.ah=IOCTL;
        regs.h.dl |= BMODE;
        regs.h.dh=0;
        regs.h.al=0x01;
        intdos(&regs,&regs);
        if (regs.x.cflag == 1) {close(handle); return(-1);}

        /* Next do function 3 to activate device driver */
        regs.h.ah=IOCTL;
        regs.h.al = 0x03;
        regs.x.dx=(int) &ddata;

        regs.x.cx=2;
        intdos(&regs,&regs);
        if (regs.x.cflag == 1) {close(handle); return(-1);}

        /* LSM 7/18/86 code to let Bristol know that rollup is attached */
        if (sendit(H_WRITE,"{ _ROLLUP=TRUE;IOCTL(ROLLUP,3,@'0A00@'X); }")) {
                close(handle); return(-1);
                }

        /* Else all went well */
        return(0);
        }

/****************************************************************************/
/* Procedure Sendit -                                                       */
/*      Starts the up or download by sending the host transaction.          */
/*      Transaction points to either "400" or "401", depending on if an     */
/*      up or download is being done.  Filename points to the file name     */
/*      to be used by the controller.                                       */
/****************************************************************************/
static int sendit(transaction,filename)
        char *transaction;              /* Upload or download transaction */
        char *filename;                 /* File to be up or downloaded    */
        {
        /* Build string to be sent to controller
                0,seq,transaction,text
                                                */
        /* change seq number */
        seq++;

        strcpy(xmit,"0,");             /* Begin string */
        itoa(seq,xmit+2,10);           /* Add seq number */
        strcat(xmit,",");              /* Add comma */
        strcat(xmit,transaction);      /* Add transaction */
        strcat(xmit,",'");             /* Add ,'   */

        strcat(xmit,filename);         /* Add msg */
        strcat(xmit,"'");              /* Add final ' */

        /* Send request.  Tracewrite exits immediately if error */
        tracewrite(strlen(xmit),7,0);

        return(0);
        }

/****************************************************************************/
/* Procedure Getrecv -                                                      */
/*      Gets a response from the controller and places it in recv.          */
/*      Returns:                                                            */
/*              -2 : EOT received                                           */
/*              -1 : Device Driver Error                                    */
/*               0 : No characters read --> Timeout                         */
/*              >0 : Number of characters read                              */
/*                                                                          */
/****************************************************************************/
static int getrecv(timeout)
        int timeout;
        {
        long starttime;
        long endtime;
        int numread;
        char *p;
        int i;

        time(&starttime);

        /* Get recv from controller */
        for (endtime=starttime;!numinbuffer() && endtime-starttime < timeout;time(&endtime)) ;

        if (numinbuffer() == -1) {
                /* End of File received */
                read(handle,recv,1);

                if (traceflag) {
                        fprintf(stdprn,"RECV: EOF\n");
                        fflush(stdprn);
                        }
                return(-2);
                }

        if ((numread=read(handle,recv,BUFSIZE)) > 0) {
                if (traceflag) {
                        fprintf(stdprn,"RECV: ");
                        for (i=0,p=recv;i<numread;i++) fputc(*p++,stdprn);
                        fputc('\n',stdprn);
                        fflush(stdprn);
                        }
                }

        return(numread);
        }



/****************************************************************************/
/* Procedure Getit -                                                        */
/*      This procedure is called whenever we expect a line from the         */
/*      controller.  It interprets the line, and returns a return code      */
/*      based on what it thinks the line means.                             */
/*                                                                          */
/*      Returns:                                                            */
/*      0    ==> A host message that begins with a 0 or 1 and the           */
/*               sequence number matches (i.e. a final response msg).       */
/*      3    ==> timeout error                                              */
/*      4    ==> read error                                                 */
/*      5    ==> command rejected                                           */
/*      6    ==> invalid seq number (compares to seq number used)           */
/*      7    ==> unsolicited write (i.e. an EXCP message )                  */
/*      8    ==> free formated message (i.e. one that is not in the         */
/*                                      form of a host message).            */
/*      11 + ==> An error to the host request.  This is the return code     */
/*               that the host processor gives.                             */
/****************************************************************************/
static int getit()
        {
        int answer;                     /* Points to answer within msg      */
        int numread=getrecv(timeout);   /* The number of characters read    */
        int newseq;                     /* the seq number of a received msg */

        /* See if device driver error */
        if (numread < 0) return(4);

        /* See if timeout */
        if (numread == 0) return(3);

        /* If less than a full buffer read, blank out top position */
        if (numread<BUFSIZE) recv[numread]='\0';

        /* See if a free formated message (i.e. a SENDFILE, READFILE,) */
        if (!isdigit(*recv) || recv[1] != ' ') return(8);

        /* See if command rejected */
        if (recv[0]=='3') return(5);

        /* Find where answer begins */
        for (answer=1;recv[answer]==' ';answer++);
        while (recv[++answer] != ' ') ;
        while (recv[++answer]==' ');

        /* Fetch the seq number */
        newseq=atoi(recv+2);

        /* See if an unsolicited write */
        if (*recv=='0' && atoi(recv+answer)==222) return(7);

        /* See if seq number matches */
        if (seq != newseq ) return(6);

        /* See if error occurred */
        if (recv[0]=='2') return(10+recv[answer]-'0');

        return(0);
        }

/****************************************************************************/
/* Procedure Getfn -                                                        */
/*      Finds where the filename begins in a string.  It does so by         */
/*      searching from the end of the string backwards, until the first     */
/*      \ or : is encountered.                                              */
/****************************************************************************/
char *getfn(pathname)
        char *pathname;
        {
        char *p;

        /* find where filename in pathname begins */
        for (p=pathname+strlen(pathname);p > pathname && *p != ':' && *p != '\\';p--) ;

        /* If we're pointing to a character, must be : or \, so return character
           after the : or \.   */
        return( (*p==':' || *p=='\\') ? p+1 : p);
        }




/****************************************************************************/
/* Procedure Commandline -                                                  */
/*      Gets arguments from DOS.                                            */
/****************************************************************************/
commandline(argc,argv)
        int argc;
        char *argv[];
        {
        char *p;                /* Points to start of parameter in argument */
        char *q;                /* Points to end of parameter in argument   */
        int i;                  /* Miscellaneous loop variable              */

        /* First check that at least one argument is given */
        if (argc < 2) goodbye(0);

        /* Initialize device */
        device[0]='\0';

        /* Loop through each argument, looking for /D:, /T: or /P */
        /* If one is found, extract value and replace with blanks */
        for (i=1;i<argc;i++)
                /* Loop over all parameters in the arg */
                while ((p=strchr(argv[i],'/')) != NULL) {
                        /* Branch accordingly */
                        switch (*(p+1)) {
                        case 'P':
                        case 'p':       /* Printing desired */
                                        traceflag=1;
                                        q=p+1;

                                        /* Make sure noting follows the P */
                                        if (*(q+1) && *(q+1)!='/') goodbye(8);
                                        break;
                        case 'T':
                        case 't':       /* New timeout requested */
                                        /* Make sure : follows T */
                                        if (*(p+2) != ':') goodbye(8);

                                        /* Make sure each character is a digit*/
                                        for (q=p+3;*q && isdigit(*q);q++) ;

                                        /* Ending character better be \0 or / */
                                        /* Decrement q so that it points to the
                                           real end of the parameter          */
                                        /* make sure that it doesn't overwite
                                           string terminator - ME 4/30/86 */
                                        if (*q && *(q-1) != '/') goodbye(8);
                                        q--;


                                        /* Lets fetch the timeout */
                                        timeout=atoi(p+3);

                                        /* make sure that the timeout is at
                                           least the default amount */
                                        if (timeout < DEF_TIMEOUT)
                                                timeout=DEF_TIMEOUT;
                                        break;
                        case 'D':
                        case 'd':       /* device driver requested */
                                        /* Make sure : follows D */
                                        if (*(p+2) != ':') goodbye(8);

                                        /* Copy up to 8 characters of device */
                                        strncpy(device,p+3,8);
                                        device[8]='\0';

                                        /* If we copied a /, then remove it  */
                                        if ((q=strchr(device,'/')) != NULL)
                                                *q='\0';

                                        /* Now point q to end of parm */
                                        for (q=p+3;*q && *q != '/';q++) ;

                                        q--;
                                        break;
                        default:        goodbye(8);
                        }

                        /* Blank out parameter */
                        while (p <= q) *p++ = ' ';
                        }
        }


/****************************************************************************/
/* Procedure Main -                                                         */
/*      The entry point from DOS.                                           */
/****************************************************************************/
main(argc,argv)
        int argc;
        char *argv[];
        {
        int i;                  /* Used as a miscellaneous loop variable    */
        char *p;                /* Points to start of first filename        */
        char *q;                /* Points to start of second filename       */
        int download=0;         /* 1 indicates a download, 2 an upload      */
        long starttime;         /* The number of seconds since 1-1-80       */
        char *rollupfile;       /* The file to be updownload on the rollup  */
        char *amlfile;          /* The file to be updownload on controller  */
        int loopflag=1;         /* Loop until controller responds flag      */
        int rolluphandle;       /* The file handle of the file on the rollup*/
        int count;              /* the number of characters read            */
        int earlyresponse;      /* Indicates if the controller has resp'd   */
        int rc;                 /* return code from getit                   */
        struct stat buffer;      /* needed stat C function call */


        commandline(argc,argv);

        /* Set sequence number to be a random value based on the time */
        time(&starttime);
        seq=abs((int) starttime);

        /* Check to make sure we have a legal device name */
        if (!device[0]) strcpy(device,"ROLLUP1");

        /* Open device driver */
        if (openport(device)) goodbye(5);

        /* Now point p to the first file, q to the second file.  Strip
           off leading and trailing blanks */
        for (p=argv[1];*p && isspace(*p);p++) ;
        if (strchr(p,' ') != NULL) *(strchr(p,' ')) = '\0';

        for (q=argv[2];*q && isspace(*q);q++) ;
        if (strchr(q,' ') != NULL) *(strchr(q,' ')) = '\0';

        /* Check to make sure first file is not empty */
        if (!(*p)) goodbye(0);

        /* Lets see if an upload (i.e. R: on first file) */
        if ((*p == 'R' || *p == 'r') && *(p+1) == ':') {
                download=2;

                /* if there is no file after the R: then return msg that
                   2 files must be given   LSM 4/30/85 */
                if (*(p+2)=='\0') goodbye(0);

                /* Point controller file after the R: */
                amlfile=p+2;

                /* If we haven't got a rollup file, use the first file with
                   the path removed */
                if (!(*q)) q=getfn(amlfile);

                /* Point rollupfile to original argument or file within first
                   filespec */
                rollupfile=q;

                }

        /* Check to make sure second file is not empty.  If we requested an
           upload and no file was given, q has been changed to point to
           the filename within the first filespec.  If a download is
           requested, then q is unchanged from the user's second parameter,
           and may be empty */
        if (!(*q)) goodbye(0);

        /* Lets see if a download (i.e. R: on second file) */
        if ((*q == 'R' || *q == 'r') && *(q+1) == ':')
                /* Check to make sure an upload isn't also requested */
                if (download) goodbye(2);
                        else {
                        download=1;
                        rollupfile=p;
                        amlfile=q+2;

                        /* Allow R: for download.  If no filename is given
                           following the R:, use filename from first file */
                        if (!(*amlfile)) amlfile=getfn(rollupfile);

                        }

        /* Make sure at least one R: appears */
        if (!download) goodbye(1);

        if (download==1) {
                /* Open file to be downloaded */
                if ((rolluphandle=open(rollupfile,O_RDONLY|O_BINARY)) == -1)
                        goodbye(3);

                /* make sure not to download file of zero length */
                stat(rollupfile,&buffer);
                if (buffer.st_size==0)
                        /* send back msg that unable to open source file */
                        goodbye(3);

                /* Print starting message */
                printf("Now downloading %s to controller disk as %s\n",rollupfile,amlfile);
                }
        else    {
                /* Open the file on the rollup, destoying any previous file */
                if ((rolluphandle=open(rollupfile,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,
                                                 S_IREAD|S_IWRITE)) == -1)
                        goodbye(4);

                /* Print starting message */
                printf("Now uploading %s from controller disk as %s\n",amlfile,rollupfile);
                }

        /* Lets first flush any characters in the receive buffer */
        while (getrecv(1));

        /* Let's start the upload/download transaction */
        sendit((download==1) ? H_DOWNLOAD : H_UPLOAD,amlfile);

        /* Let's get response.  Must be a freeformated SENDFILE or READFILE
           message. */
        while (loopflag) switch(rc=getit()) {
                case 3:
                                /* if doing a upload, make sure that file is
                                erased or else have file of 0 bytes created */
                                if (download==2)  {
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        }
                                goodbye(7);
                case 8:         /* end looping if correct response */
                                if (!recv[1])
                                        if ((download==1 && *recv == SENDFILE)
                                         || (download==2 && *recv == READFILE))
                                                loopflag=0;
                                /* Keep looping if incorrect controller response */
                case 4:
                case 0:
                case 5:
                case 6:         break;

                case 11:        /* can't do down/upload while pgm runnung */
                                /* if doing a upload, make sure that file is
                                 erased or else have file of 0 bytes created */
                                if (download==2)  {
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        }
                                goodbye(9);
                default:        /* If unsolicited write or an exception occurred
                                to the host request, display the exception */
                                /* if doing a upload, make sure that file is
                                 erased or else have file of 0 bytes created */
                                if (download==2)  {
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        }
                                goodbye(12);
                }

        /* We've now received the SENDFILE/READFILE request */
        /* Send the opposite message to the controller */
        *xmit = (download==1) ? READFILE : SENDFILE;

        /* if we get an error, kill the program */
        tracewrite(1,7,0);

        /* Lets do the download or the upload! */
        if (download==1) {
                /* Perform download */
                /* Keep looping until EOF */
                while ((count=read(rolluphandle,xmit,BUFSIZE/2)) > 0 && !(earlyresponse=numinbuffer())) {
                        tracewrite(count,7,1);
                        }

                /* Finish the protocol if the controller has not responded
                   early */
                if (!earlyresponse) {
                        /* Finally reached EOF or read error occurred */
                        /* Send end of file marker */
                        xmiteof();

                        /* Send GOODFILE or BADFILE based on if a read error occurred */
                        *xmit=(count == 0) ? GOODFILE : BADFILE;

                        /* Wait for space in XMIT buffer of device driver */
                        tracewrite(1,7,0);
                        }

                /* Let's get response.  Must be a freeformated FILESAVED msg */
                loopflag=1;
                while (loopflag) switch(rc=getit()) {
                        case 3:         goodbye(7);
                        case 8:         /* See if correct response */
                                        if (!recv[1] && (*recv==FILESAVED ||
                                                         *recv==NOTSAVED))
                                                loopflag=0;
                                        /* Keep looping if incorrect controller response */
                        case 4:
                        case 0:
                        case 5:
                        case 6:         break;

                        case 11:        goodbye(9);

                        default:        /* If unsolicited write or an exception occurred
                                        to the host request, display the exception */
                                        goodbye(12);
                        }
                }
        else    {
                /* Perform upload */
                /* First read data ready command */
                loopflag=1;
                while (loopflag) switch(rc=getit()) {
                        case 3:
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        goodbye(7);
                        case 8:         /* end looping if correct response */
                                        if (!recv[1] && *recv == DATAREADY)
                                                loopflag=0;
                                        /* Keep looping if incorrect controller response */
                        case 4:
                        case 0:
                        case 5:
                        case 6:         break;

                        case 11:
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        goodbye(9);
                        default:        /* If unsolicited write or an exception occurred
                                        to the host request, display the exception */
                                        /* Close and remove file */
                                        close(rolluphandle);
                                        unlink(rollupfile);
                                        goodbye(12);
                        }

                /* Now loop, reading each record from the device driver */
                while (1) {
                        /* Read read record.  Returns -1 if error, -2 if
                           EOF, 0 if timeout, or >0 if data */
                        count=getrecv(timeout);

                        /* If unable to write received record, or timeout
                           or device driver error, tell controller that file
                           was not saved and end the program */
                        if (count == -1 || count ==0 ||
                           (count > 0 && write(rolluphandle,recv,count) < count)) {
                                /* Tell controller that file wasn't saved */
                                *xmit=NOTSAVED;

                                /* Close and remove file */
                                close(rolluphandle);
                                unlink(rollupfile);

                                /* Set return code to make most sense */
                                if (count==-1) rc=6;
                                        else if (count==0) rc=7;
                                                else rc=11;

                                /* Wait for space in XMIT buffer of device driver */
                                tracewrite(1,rc,0);

                                /* exit the program */
                                goodbye(rc);
                                }

                        /* if count = -2, then must be at eof! */
                        if (count == -2) break;
                        }
                /* Have now read EOF, now read GOODFILE/BADFILE */
                loopflag=1;
                while (loopflag) switch(rc=getit()) {
                        case 3:         goodbye(7);
                        case 8:         /* end looping if correct response */
                                        if (!recv[1] && (*recv == GOODFILE ||
                                                         *recv == BADFILE))
                                                loopflag=0;
                                        /* Keep looping if incorrect controller response */
                        case 4:
                        case 0:
                        case 5:
                        case 6:         break;

                        case 11:        goodbye(9);

                        default:        /* If unsolicited write or an exception occurred
                                        to the host request, display the exception */
                                        goodbye(12);
                        }

                /* Send FILESAVED or NOTSAVED based on if a GOODFILE or BADFILE */
                *xmit=(*recv==GOODFILE) ? FILESAVED : NOTSAVED;

                /* Close and remove file if NOTSAVED*/
                if (close(rolluphandle)) *xmit=NOTSAVED;

                if (*xmit==NOTSAVED) unlink(rollupfile);

                /* Wait for space in XMIT buffer of device driver */
                tracewrite(1,(*xmit==NOTSAVED) ? 11 : 7,0);
                }



        /* Now done with separate upload/download code, see if correct resp */
        loopflag=1;
        while (loopflag) switch(rc=getit()) {
                case 3:         goodbye(7);

                case 0:         loopflag=0;

                case 6:         /* If incorrect seq # or free format loop */
                case 4:         /* Or if incorrect controller response!   */
                case 5:
                case 8:         break;

                case 11:        goodbye(9);

                default:        /* If unsolicited write or an exception occurred
                                to the host request, display the exception */
                                goodbye(12);
                }
         /* LSM 7/18/86 code to let Bristol know that rollup is not attached */
        sendit(H_WRITE,"{ IOCTL(ROLLUP,3,@'0B00@'X);_ROLLUP=FALSE; }");

        exit(0);
        }
